from contessa.db import Connector
from contessa.models import DQBase


def test_upsert_qualitycheck(conn: Connector):
    from sqlalchemy import Column, DateTime, text, UniqueConstraint
    from sqlalchemy.dialects.postgresql import TEXT, INTEGER, BIGINT

    class A(DQBase):
        id = Column(BIGINT, primary_key=True)
        name = Column(TEXT, nullable=False)
        price = Column(INTEGER)
        created_at = Column(
            DateTime(timezone=True),
            server_default=text("NOW()"),
            nullable=False,
            index=True,
        )

        __tablename__ = "my_table"
        __table_args__ = (UniqueConstraint("name", name=f"unique_constraint_test",),)

    conn.ensure_table(A.__table__)
    instance = A(name="hello", price=13)

    conn.upsert(
        objs=[instance,]
    )

    # check if inserted
    s = conn.make_session()
    row = s.query(A.__table__).all()
    s.expunge_all()
    s.commit()
    assert len(row) == 1
    assert row[0].price == 13

    # change data and insert again - should upsert
    instance.price = 42
    conn.upsert(
        objs=[instance,]
    )

    row = s.query(A.__table__).all()
    s.expunge_all()
    s.commit()
    assert len(row) == 1
    assert row[0].price == 42

    s.close()



from contessa.db import Connector
from contessa.models import DQBase
from test.conftest import FakedDatetime
from test.integration.conftest import TEST_DB_URI
import unittest
from unittest import mock

from contessa import ConsistencyChecker


class TestDataQualityOperator(unittest.TestCase):
    def setUp(self):
        """
        Init a temporary table with some data.
        """
        self.left_table_name = "raw_booking"
        self.right_table_name = "booking"
        self.result_table_name = "booking"
        self.ts_nodash = (
            FakedDatetime.now().isoformat().replace("-", "").replace(":", "")
        )
        self.now = FakedDatetime.now()

        sql = [
            "DROP SCHEMA if exists tmp CASCADE;",
            "DROP SCHEMA if exists data_quality CASCADE;",
            "CREATE SCHEMA IF NOT EXISTS tmp;",
            "CREATE SCHEMA IF NOT EXISTS data_quality;",
            "CREATE SCHEMA IF NOT EXISTS hello;",
            f"""
                    CREATE TABLE IF NOT EXISTS tmp.{self.left_table_name}(
                      id SERIAL PRIMARY KEY,
                      src text,
                      dst text,
                      price int,
                      turnover_after_refunds double precision,
                      initial_price double precision,
                      created_at timestamptz
                    )
                """,
            f"""
                    CREATE TABLE IF NOT EXISTS hello.{self.right_table_name}(
                      id SERIAL PRIMARY KEY,
                      src text,
                      dst text,
                      price int,
                      turnover_after_refunds double precision,
                      initial_price double precision,
                      created_at timestamptz
                    )
                """,
            f"""
                INSERT INTO tmp.{self.left_table_name}
                    (src, dst, price, turnover_after_refunds, initial_price, created_at)
                VALUES
                    ('BTS', NULL, 1, 100, 11, '2018-09-12T11:50:00'),
                    (NULL, 'PEK', 33, 1.1, 13, '2018-01-12T15:50:00'),
                    ('VIE', 'JFK', 4, 5.5, 23.4, '2018-09-11T11:50:00'),
                    ('VIE', 'VIE', 4, 0.0, 0.0, '2018-09-11T11:50:00')
                """,
            f"""
                INSERT INTO hello.{self.right_table_name}
                    (src, dst, price, turnover_after_refunds, initial_price, created_at)
                VALUES
                    ('BTS', NULL, 1, 100, 11, '2018-09-12T11:50:00'),
                    (NULL, 'PEK', 33, 1.1, 13, '2018-01-12T15:50:00'),
                    ('VIE', 'JFK', 4, 5.5, 23.4, '2018-09-11T11:50:00')
            """,
        ]
        self.conn = Connector(TEST_DB_URI)
        for s in sql:
            self.conn.execute(s)

        self.consistency_checker = ConsistencyChecker(TEST_DB_URI)

    def tearDown(self):
        """
        Drop all created tables.
        """
        self.conn.execute(f"DROP schema tmp CASCADE;")
        self.conn.execute(f"DROP schema data_quality CASCADE;")
        self.conn.execute(f"DROP schema hello CASCADE;")
        DQBase.metadata.clear()

    @mock.patch("contessa.executor.datetime", FakedDatetime)
    def test_execute_consistency_false(self):
        self.consistency_checker.run(
            self.consistency_checker.COUNT,
            left_check_table={"schema_name": "tmp", "table_name": self.left_table_name},
            right_check_table={
                "schema_name": "hello",
                "table_name": self.right_table_name,
            },
            result_table={
                "schema_name": "data_quality",
                "table_name": self.result_table_name,
            },
            context={"task_ts": self.now},
        )
        rows = self.conn.get_pandas_df(
            f"""
            SELECT * from data_quality.consistency_check_{self.result_table_name}
            order by created_at
        """
        )
        self.assertEqual(rows.iloc[0]["status"], "invalid")

    @mock.patch("contessa.executor.datetime", FakedDatetime)
    def test_execute_consistency_true(self):
        # add missing record to the right table
        self.conn.execute(
            f"""
            INSERT INTO hello.{self.right_table_name}
                (src, dst, price, turnover_after_refunds, initial_price, created_at)
            VALUES
                ('VIE', 'VIE', 4, 0.0, 0.0, '2018-09-11T11:50:00')
        """
        )

        self.consistency_checker.run(
            self.consistency_checker.COUNT,
            left_check_table={"schema_name": "tmp", "table_name": self.left_table_name},
            right_check_table={
                "schema_name": "hello",
                "table_name": self.right_table_name,
            },
            result_table={
                "schema_name": "data_quality",
                "table_name": self.result_table_name,
            },
            context={"task_ts": self.now},
        )

        rows = self.conn.get_pandas_df(
            f"""
            SELECT * from data_quality.consistency_check_{self.result_table_name}
            order by created_at
        """
        )
        self.assertEqual(rows.iloc[0]["status"], "valid")

    @mock.patch("contessa.executor.datetime", FakedDatetime)
    def test_execute_consistency_diff(self):
        """ Test scenario where table diff is done on tables with different column order."""
        self.inconsistent_colums_left = "user"
        self.inconsistent_colums_right = "user_inconsistent"
        self.conn = Connector(TEST_DB_URI)
        self.consistency_checker = ConsistencyChecker(TEST_DB_URI)

        sql = [
            f"""
                CREATE TABLE IF NOT EXISTS tmp.{self.inconsistent_colums_left}(
                  id SERIAL PRIMARY KEY,
                  name text
                )
            """,
            f"""
                CREATE TABLE IF NOT EXISTS tmp.{self.inconsistent_colums_right}(
                  name text,
                  id SERIAL PRIMARY KEY
                )
            """,
            f"""
                INSERT INTO tmp.{self.inconsistent_colums_left}
                VALUES
                    (1, 'John Doe')
                """,
            f"""
                INSERT INTO tmp.{self.inconsistent_colums_right}
                VALUES
                    ('John Doe', 1)
            """,
        ]

        for s in sql:
            self.conn.execute(s)

        self.consistency_checker.run(
            self.consistency_checker.DIFF,
            left_check_table={
                "schema_name": "tmp",
                "table_name": self.inconsistent_colums_left,
            },
            right_check_table={
                "schema_name": "tmp",
                "table_name": self.inconsistent_colums_right,
            },
            result_table={
                "schema_name": "data_quality",
                "table_name": self.result_table_name,
            },
            context={"task_ts": self.now},
        )

        rows = self.conn.get_pandas_df(
            f"""
                SELECT * from data_quality.consistency_check_{self.result_table_name}
                order by created_at
            """
        )

        self.assertEqual(rows.iloc[0]["status"], "valid")



from datetime import timedelta

from contessa.settings import TIME_FILTER_DEFAULT


def compose_where_time_filter(time_filter, task_ts):
    days = 30
    filters = []
    result = []
    if time_filter != TIME_FILTER_DEFAULT:
        if isinstance(time_filter, str):
            filters.append({"column": time_filter, "days": days})
        elif isinstance(time_filter, list):
            for each in time_filter:
                filters.append(
                    {"column": each["column"], "days": each.get("days", days)}
                )

        for each in filters:
            present = task_ts.strftime("%Y-%m-%d %H:%M:%S UTC")
            past = (task_ts - timedelta(days=each["days"])).strftime(
                "%Y-%m-%d %H:%M:%S UTC"
            )
            result.append(
                f"""({each["column"]} >= '{past}'::timestamptz AND {each["column"]} < '{present}'::timestamptz)"""
            )

    return " OR ".join(result)



import pytest

import contessa
from contessa import ContessaRunner


def test_attribute_in_custom_sql_rule():
    r = [
        {
            "name": "test_rule_name",
            "type": contessa.SQL,
            "sql": "select 1",
            "column": "col_name",
            "description": "some description",
        },
        {
            "name": "test_gt_rule",
            "type": contessa.GT,
            "value": 3,
            "column": "mycol",
            "description": "desc",
        },
    ]
    rules = ContessaRunner.build_rules(r)
    assert rules[0].attribute == "col_name"


def test_no_attribute_in_custom_sql_rule():
    r = [
        {
            "name": "test_rule_name",
            "type": contessa.SQL,
            "sql": "select 1",
            "description": "some description",
        }
    ]
    with pytest.raises(TypeError):
        ContessaRunner.build_rules(r)



import unittest

from contessa.db import Connector
from test.integration.conftest import TEST_DB_URI
from contessa.models import DQBase
from contessa.migration import MigrationsResolver

DATA_QUALITY_SCHEMA = "data_quality_test"
MIGRATION_TABLE = "contessa_alembic_version"
DATA_QUALITY_TABLE_1 = "quality_check_example_table"
DATA_QUALITY_TABLE_2 = "quality_check_another_table"
SQLALCHEMY_URL = "postgresql://postgres:postgres@postgres:5432/test_db"


def get_quality_table_creation_script(schema, table_name):
    return f"""create table {schema}.{table_name}
                    (
                        attribute text,
                        rule_name text,
                        rule_description text,
                        total_records integer,
                        failed integer,
                        median_30_day_failed double precision,
                        failed_percentage double precision,
                        passed integer,
                        median_30_day_passed double precision,
                        passed_percentage double precision,
                        status text,
                        time_filter text,
                        task_ts timestamp with time zone not null,
                        created_at timestamp with time zone default now() not null,
                        id bigserial not null
                            constraint {table_name}_pkey
                                primary key,
                        constraint {table_name}_unique_quality_check
                            unique (attribute, rule_name, task_ts, time_filter)
                    );
                """


class TestMigrationsResolverInit(unittest.TestCase):
    def setUp(self):
        """
        Init a temporary table with some data.
        """

        sql = [
            f"DROP SCHEMA IF EXISTS {DATA_QUALITY_SCHEMA} CASCADE;"
            f"CREATE SCHEMA IF NOT EXISTS {DATA_QUALITY_SCHEMA};",
            get_quality_table_creation_script(
                DATA_QUALITY_SCHEMA, DATA_QUALITY_TABLE_1
            ),
            get_quality_table_creation_script(
                DATA_QUALITY_SCHEMA, DATA_QUALITY_TABLE_2
            ),
        ]
        self.conn = Connector(TEST_DB_URI)
        for s in sql:
            self.conn.execute(s)

    def tearDown(self):
        """
        Drop all created tables.
        """
        self.conn.execute(f"DROP schema {DATA_QUALITY_SCHEMA} CASCADE;")
        DQBase.metadata.clear()

    def test_migration_table_exists_init(self):
        versions_migrations = {"0.1.4": "0.1.4-hash", "0.1.5": "0.1.5-hash"}

        m = MigrationsResolver(
            versions_migrations, "0.1.4", SQLALCHEMY_URL, DATA_QUALITY_SCHEMA
        )
        migration_table_exists = m.migrations_table_exists()

        assert migration_table_exists is False

    def test_get_current_migration_init(self):
        versions_migrations = {"0.1.4": "0.1.4-hash", "0.1.5": "0.1.5-hash"}

        m = MigrationsResolver(
            versions_migrations, "0.1.4", SQLALCHEMY_URL, DATA_QUALITY_SCHEMA
        )
        current = m.get_applied_migration()

        assert current is None

    def test_is_on_head_init(self):
        versions_migrations = {"0.1.4": "0.1.4-hash", "0.1.5": "0.1.5-hash"}

        m = MigrationsResolver(
            versions_migrations, "0.1.4", SQLALCHEMY_URL, DATA_QUALITY_SCHEMA
        )
        is_on_head = m.is_on_head()

        is_on_head is False

    def test_get_migrations_to_head__is_before_init(self):
        versions_migrations = {
            "0.1.2": "0.1.2-hash",
            "0.1.3": "0.1.3-hash",
            "0.1.4": "0.1.4-hash",
            "0.1.5": "0.1.5-hash",
            "0.1.6": "0.1.6-hash",
            "0.1.7": "0.1.7-hash",
        }

        m = MigrationsResolver(
            versions_migrations, "0.1.7", SQLALCHEMY_URL, DATA_QUALITY_SCHEMA
        )
        migrations = m.get_migration_to_head()
        assert migrations[0] is "upgrade"
        assert migrations[1] is "0.1.7-hash"


class TestMigrationsResolver(unittest.TestCase):
    def setUp(self):
        """
        Init a temporary table with some data.
        """

        sql = [
            f"DROP SCHEMA IF EXISTS {DATA_QUALITY_SCHEMA} CASCADE;"
            f"CREATE SCHEMA IF NOT EXISTS {DATA_QUALITY_SCHEMA};",
            get_quality_table_creation_script(
                DATA_QUALITY_SCHEMA, DATA_QUALITY_TABLE_1
            ),
            get_quality_table_creation_script(
                DATA_QUALITY_SCHEMA, DATA_QUALITY_TABLE_2
            ),
            f"""
                    create table {DATA_QUALITY_SCHEMA}.{MIGRATION_TABLE}
                        (
                            version_num varchar(32) not null
                                constraint alembic_version_pkc
                                    primary key
                        );
                        INSERT INTO {DATA_QUALITY_SCHEMA}.{MIGRATION_TABLE} (version_num) VALUES ('0.1.4-hash');
                    """,
        ]
        self.conn = Connector(TEST_DB_URI)
        for s in sql:
            self.conn.execute(s)

    def tearDown(self):
        """
        Drop all created tables.
        """
        # self.conn.execute(f"DROP schema data_quality_test CASCADE;")
        DQBase.metadata.clear()

    def test_schema_exists(self):
        versions_migrations = {"0.1.4": "0.1.4-hash", "0.1.5": "0.1.5-hash"}

        m = MigrationsResolver(
            versions_migrations, "0.1.4", SQLALCHEMY_URL, DATA_QUALITY_SCHEMA
        )
        schema_exists = m.schema_exists()

        assert schema_exists

        m = MigrationsResolver(
            versions_migrations, "0.1.4", SQLALCHEMY_URL, "not_existing_schema"
        )
        schema_exists = m.schema_exists()

        assert schema_exists is False

    def test_migration_table_exists(self):
        versions_migrations = {"0.1.4": "0.1.4-hash", "0.1.5": "0.1.5-hash"}

        m = MigrationsResolver(
            versions_migrations, "0.1.4", SQLALCHEMY_URL, DATA_QUALITY_SCHEMA
        )
        migration_table_exists = m.migrations_table_exists()

        assert migration_table_exists

    def test_get_current_migration(self):
        versions_migrations = {"0.1.4": "0.1.4-hash", "0.1.5": "0.1.5-hash"}

        m = MigrationsResolver(
            versions_migrations, "0.1.4", SQLALCHEMY_URL, DATA_QUALITY_SCHEMA
        )
        current = m.get_applied_migration()

        assert current == "0.1.4-hash"

    def test_is_on_head(self):
        versions_migrations = {"0.1.4": "0.1.4-hash", "0.1.5": "0.1.5-hash"}

        m = MigrationsResolver(
            versions_migrations, "0.1.4", SQLALCHEMY_URL, DATA_QUALITY_SCHEMA
        )
        is_on_head = m.is_on_head()

        assert is_on_head

    def test_is_on_head_no_on_head(self):
        versions_migrations = {"0.1.4": "0.1.4-hash", "0.1.5": "0.1.5-hash"}

        m = MigrationsResolver(
            versions_migrations, "0.1.5", SQLALCHEMY_URL, DATA_QUALITY_SCHEMA
        )
        is_on_head = m.is_on_head()

        assert is_on_head is False

    def test_is_on_head_with_fallback(self):
        versions_migrations = {"0.1.4": "0.1.4-hash", "0.1.6": "0.1.6-hash"}

        m = MigrationsResolver(
            versions_migrations, "0.1.5", SQLALCHEMY_URL, DATA_QUALITY_SCHEMA
        )
        is_on_head = m.is_on_head()

        assert is_on_head

    def test_get_migrations_to_head__already_on_head(self):
        versions_migrations = {"0.1.4": "0.1.4-hash", "0.1.5": "0.1.5-hash"}

        m = MigrationsResolver(
            versions_migrations, "0.1.4", SQLALCHEMY_URL, DATA_QUALITY_SCHEMA
        )
        migrations = m.get_migration_to_head()
        assert migrations is None

    def test_get_migrations_to_head__package_greather_than_map_max(self):
        versions_migrations = {"0.1.4": "0.1.4-hash", "0.1.5": "0.1.5-hash"}

        m = MigrationsResolver(
            versions_migrations, "0.1.6", SQLALCHEMY_URL, DATA_QUALITY_SCHEMA
        )
        migrations = m.get_migration_to_head()
        assert migrations[0] is "upgrade"
        assert migrations[1] is "0.1.5-hash"

    def test_get_migrations_to_head__is_down_from_head(self):
        versions_migrations = {
            "0.1.2": "0.1.2-hash",
            "0.1.3": "0.1.3-hash",
            "0.1.4": "0.1.4-hash",
            "0.1.5": "0.1.5-hash",
            "0.1.6": "0.1.6-hash",
            "0.1.7": "0.1.7-hash",
        }

        m = MigrationsResolver(
            versions_migrations, "0.1.7", SQLALCHEMY_URL, DATA_QUALITY_SCHEMA
        )
        migrations = m.get_migration_to_head()
        assert migrations[0] is "upgrade"
        assert migrations[1] is "0.1.7-hash"

    def test_get_migrations_to_head__is_down_from_head_with_fallback(self):
        versions_migrations = {
            "0.1.2": "0.1.2-hash",
            "0.1.3": "0.1.3-hash",
            "0.1.4": "0.1.4-hash",
            "0.1.5": "0.1.5-hash",
            "0.1.8": "0.1.8-hash",
            "0.1.9": "0.1.9-hash",
        }

        m = MigrationsResolver(
            versions_migrations, "0.1.7", SQLALCHEMY_URL, DATA_QUALITY_SCHEMA
        )
        migrations = m.get_migration_to_head()
        assert migrations[0] is "upgrade"
        assert migrations[1] is "0.1.5-hash"

    def test_get_migrations_to_head__is_up_from_head(self):
        versions_migrations = {
            "0.1.2": "0.1.2-hash",
            "0.1.3": "0.1.3-hash",
            "0.1.4": "0.1.4-hash",
            "0.1.5": "0.1.5-hash",
            "0.1.6": "0.1.6-hash",
            "0.1.7": "0.1.7-hash",
        }

        m = MigrationsResolver(
            versions_migrations, "0.1.2", SQLALCHEMY_URL, DATA_QUALITY_SCHEMA
        )
        migrations = m.get_migration_to_head()
        assert migrations[0] is "downgrade"
        assert migrations[1] is "0.1.2-hash"

    def test_get_migrations_to_head__is_up_from_head_with_fallback(self):
        versions_migrations = {
            "0.1.1": "0.1.1-hash",
            "0.1.3": "0.1.3-hash",
            "0.1.4": "0.1.4-hash",
            "0.1.5": "0.1.5-hash",
            "0.1.6": "0.1.6-hash",
            "0.1.7": "0.1.7-hash",
        }

        m = MigrationsResolver(
            versions_migrations, "0.1.2", SQLALCHEMY_URL, DATA_QUALITY_SCHEMA
        )
        migrations = m.get_migration_to_head()
        assert migrations[0] is "downgrade"
        assert migrations[1] is "0.1.1-hash"



from datetime import datetime, timedelta
from statistics import median
from typing import Dict
import json

import pandas as pd
from sqlalchemy import and_, Column, DateTime, MetaData, text, UniqueConstraint
from sqlalchemy.dialects.postgresql import (
    BIGINT,
    DOUBLE_PRECISION,
    INTEGER,
    TEXT,
    TIMESTAMP,
)
from sqlalchemy.ext.declarative import (
    AbstractConcreteBase,
    declarative_base,
    declared_attr,
)

from contessa.settings import TIME_FILTER_DEFAULT
from contessa.base_rules import Rule
from contessa.db import Connector

# default schema for results is `data_quality`, but it can be overridden by passing
# ResultTable to runner. constructing concrete model for QualityCheck (that's abstract) will
# swap the schemas before creation - see `create_default_quality_check_class`
DQBase = declarative_base(metadata=MetaData(schema="data_quality"))


class QualityCheck(AbstractConcreteBase, DQBase):
    """
    Representation of abstract quality check table.
    """

    __abstract__ = True
    _table_prefix = "quality_check"

    id = Column(BIGINT, primary_key=True)
    attribute = Column(TEXT, nullable=False)
    rule_name = Column(TEXT, nullable=False)
    rule_type = Column(TEXT, nullable=False)
    rule_description = Column(TEXT)
    total_records = Column(INTEGER)

    failed = Column(INTEGER)
    median_30_day_failed = Column(DOUBLE_PRECISION)
    failed_percentage = Column(DOUBLE_PRECISION)

    passed = Column(INTEGER)
    median_30_day_passed = Column(DOUBLE_PRECISION)
    passed_percentage = Column(DOUBLE_PRECISION)

    status = Column(TEXT)
    time_filter = Column(
        TEXT,
        default=TIME_FILTER_DEFAULT,
        server_default=TIME_FILTER_DEFAULT,
        nullable=False,
    )
    task_ts = Column(TIMESTAMP(timezone=True), nullable=False, index=True)
    created_at = Column(
        DateTime(timezone=True),
        server_default=text("NOW()"),
        nullable=False,
        index=True,
    )

    @declared_attr
    def __table_args__(cls):
        """
        Concrete classes derived from this abstract one should have unique check among the columns
        that below. But the constraint needs to have unique name, therefore we are using
        @declared_attr here to construct name of the constraint using its table name.
        :return:
        """
        return (
            UniqueConstraint(
                "attribute",
                "rule_name",
                "rule_type",
                "task_ts",
                "time_filter",
                name=f"{cls.__tablename__}_unique",
            ),
        )

    def init_row(
        self, rule: Rule, results: pd.Series, conn: Connector, context: Dict = None
    ):
        """
        Count metrics we want to measure using pd.Series api and set them to quality check object.
        """
        if results.isnull().any():
            raise ValueError("In results of rule.apply can't be any Null values.")

        # todo - add to doc
        self.task_ts = context["task_ts"]
        self.attribute = rule.attribute
        self.rule_name = rule.name
        self.rule_type = rule.type
        self.rule_description = rule.description

        self.total_records = results.shape[0]
        self.failed = results[results == False].shape[0]
        self.passed = results[results == True].shape[0]

        self.set_medians(conn)

        if isinstance(rule.time_filter, str):
            self.time_filter = rule.time_filter
        else:
            self.time_filter = json.dumps(rule.time_filter)
        self.failed_percentage = self._perc(self.failed, self.total_records)
        self.passed_percentage = self._perc(self.passed, self.total_records)
        self.status = "invalid" if self.failed > 0 else "valid"

    def _perc(self, a, b):
        res = 0
        try:
            res = (a / b) * 100
        except ZeroDivisionError:
            pass
        return res

    def set_medians(self, conn: Connector, days=30):
        """
        Calculate median of passed/failed quality checks from last 30 days.
        """
        now = datetime.today().date()
        past = now - timedelta(days=days)
        cls = self.__class__

        session = conn.make_session()
        checks = (
            session.query(cls.failed, cls.passed)
            .filter(and_(cls.task_ts <= str(now), cls.task_ts >= str(past)))
            .all()
        )
        session.expunge_all()
        session.commit()
        session.close()

        failed = [ch.failed for ch in checks]
        self.median_30_day_failed = median(failed) if failed else None

        passed = [ch.passed for ch in checks]
        self.median_30_day_passed = median(passed) if passed else None

    def __repr__(self):
        return f"Rule ({self.attribute} - {self.rule_name} - {self.rule_type} - {self.task_ts})"


class ConsistencyCheck(AbstractConcreteBase, DQBase):
    """
    Representation of abstract consistency check table.
    """

    __abstract__ = True
    _table_prefix = "consistency_check"

    id = Column(BIGINT, primary_key=True)
    type = Column(TEXT, nullable=False)
    name = Column(TEXT, nullable=False)
    description = Column(TEXT)
    left_table = Column(TEXT, nullable=False)
    right_table = Column(TEXT, nullable=False)

    status = Column(TEXT)
    time_filter = Column(
        TEXT,
        nullable=False,
        default=TIME_FILTER_DEFAULT,
        server_default=TIME_FILTER_DEFAULT,
    )
    task_ts = Column(TIMESTAMP(timezone=True), nullable=False, index=True)
    created_at = Column(
        DateTime(timezone=True),
        server_default=text("NOW()"),
        nullable=False,
        index=True,
    )

    @declared_attr
    def __table_args__(cls):
        """
        Concrete classes derived from this abstract one should have unique check among the columns
        that below. But the constraint needs to have unique name, therefore we are using
        @declared_attr here to construct name of the constraint using its table name.
        :return:
        """
        return (
            UniqueConstraint(
                "type",
                "name",
                "left_table",
                "right_table",
                "task_ts",
                "time_filter",
                name=f"{cls.__tablename__}_unique",
            ),
        )

    def init_row(
        self,
        check: Dict,
        status: str,
        left_table_name: str,
        right_table_name: str,
        time_filter=None,
        context: Dict = None,
    ):
        """
        Set result to consistency check object.
        """
        self.type = check["type"]
        self.task_ts = context["task_ts"]
        self.name = check["name"]
        self.description = check["description"]
        self.left_table = left_table_name
        self.right_table = right_table_name
        if isinstance(time_filter, str):
            self.time_filter = time_filter
        else:
            self.time_filter = json.dumps(time_filter)
        self.status = status

    def __repr__(self):
        return f"Rule ({self.type} - {self.name} - {self.task_ts})"


class Table:
    def __init__(self, schema_name, table_name):
        self.schema_name = schema_name
        self.table_name = table_name

    @property
    def fullname(self):
        # todo - solve sanitization. problem - can be already sanitized
        return f"{self.schema_name}.{self.table_name}"


class ResultTable(Table):
    """
    Name will be constructed as "quality_check_{load_table_name}", e.g. "quality_check_my_table".
    Can be overridden by `table_name`.
    """

    def __init__(self, schema_name, table_name, model_cls):
        self.model_cls = model_cls
        prefix = model_cls._table_prefix
        table_name = f"{prefix}_{table_name}"
        super().__init__(schema_name, table_name)

    def to_camel_case(self, snake_str):
        components = snake_str.split("_")
        return components[0] + "".join(x.title() for x in components[1:])

    @property
    def clsname(self):
        """
        Construct a name for dynamic creation of cls for quality table.
        Should be unique in runtime.
        """
        name = super().fullname.replace(".", "_")
        camel_case = self.to_camel_case(name)
        return camel_case[0].title() + camel_case[1:]


def create_default_check_class(result_table: ResultTable):
    """
    This will construct type/class (not object) that will have special name that its prefixed
    with its table. It's better because sqlalchemy can yell somethings if we would use same class
    for generic quality checks.
    Basically, its somethings as:

        class MyTableQualityCheck(QualityCheck):
            id = ...
            ...

    But it has dynamic name - MyTable is replaced for the table we are doing quality check for.
    :return: class with dynamically created name
    """

    attributedict = {
        "__tablename__": result_table.table_name,
        "id": Column(BIGINT, primary_key=True),
        "__mapper_args__": {
            "polymorphic_identity": result_table.table_name,
            "concrete": True,
        },
    }
    cls = type(result_table.clsname, (result_table.model_cls,), attributedict)
    cls.metadata.schema = result_table.schema_name
    cls.__table__.schema = result_table.schema_name
    return cls



import os

import pytest

from contessa.db import Connector

TEST_DB_URI = os.environ.get("TEST_DB_URI")
if not TEST_DB_URI:
    raise ValueError("To run integration test set `TEST_DB_URI` env var.")

ALEMBIC_INI_PATH = os.environ.get("ALEMBIC_INI_PATH")
if not ALEMBIC_INI_PATH:
    raise ValueError("To run integration test set `ALEMBIC_INI_PATH` env var.")


@pytest.fixture()
def conn():
    h = Connector(TEST_DB_URI)

    schemas = ["tmp", "temporary", "data_quality", "raw"]
    create_queries = [f"create schema if not exists {s}" for s in schemas]
    drop_queries = [f"drop schema if exists {s} cascade" for s in schemas]

    for c in create_queries:
        h.execute(c)

    yield h

    for d in drop_queries:
        h.execute(d)



import unittest

from alembic.config import Config

import contessa.migration_runner as migration
from contessa import ConsistencyChecker
from contessa.alembic.packages_migrations import migration_map

from contessa.db import Connector
from contessa.settings import TIME_FILTER_DEFAULT

from test.conftest import FakedDatetime
from test.integration.conftest import TEST_DB_URI, ALEMBIC_INI_PATH
from contessa.models import (
    DQBase,
    QualityCheck,
    ConsistencyCheck,
    create_default_check_class,
    ResultTable,
    Table,
)

DATA_QUALITY_SCHEMA = "data_quality_test"


class MigrationTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        cls.conn = Connector(TEST_DB_URI)
        cls.alembic_cfg = Config(ALEMBIC_INI_PATH)

        migration_table_name = cls.alembic_cfg.get_main_option("version_table")
        cls.migration_table = Table(DATA_QUALITY_SCHEMA, migration_table_name)

    def migrate_to_latest(self):
        """
        Mock sqlalchemy version table to be migration 1 before latest.
        Then apply latest migration.
        """
        idx = list(migration_map.keys())[-2]
        before_last = migration_map[idx]

        self.conn.execute(
            f"create table {self.migration_table.fullname}(version_num text);"
        )
        self.conn.execute(
            f"insert into {self.migration_table.fullname}(version_num) values('{before_last}')"
        )

        last_version = list(migration_map.keys())[-1]
        try:
            migration.main(
                ["-u", TEST_DB_URI, "-s", DATA_QUALITY_SCHEMA, "-v", last_version, "-f"]
            )
        except SystemExit as e:
            print(e.args[0])

    def migrate_to(self, version):
        try:
            migration.main(
                ["-u", TEST_DB_URI, "-s", DATA_QUALITY_SCHEMA, "-v", version, "-f"]
            )
        except SystemExit as e:
            print(e.args[0])

    def fullname(self, table_name, cls):
        """
        Prefix schema name to table name.
        """
        return f"{DATA_QUALITY_SCHEMA}.{cls._table_prefix}_{table_name}"


def get_quality_table_creation_script_0_1_4(table: ResultTable):
    return f"""create table {table.fullname}
                    (
                        attribute text,
                        rule_name text,
                        rule_description text,
                        total_records integer,
                        failed integer,
                        median_30_day_failed double precision,
                        failed_percentage double precision,
                        passed integer,
                        median_30_day_passed double precision,
                        passed_percentage double precision,
                        status text,
                        time_filter text,
                        task_ts timestamp with time zone not null,
                        created_at timestamp with time zone default now() not null,
                        id bigserial not null
                            constraint {table.table_name}_pkey
                                primary key
                    );
                """


def get_quality_table_creation_script_0_1_4_invalid(table):
    return f"""create table {table.fullname}
                    (
                        attribute text,
                        rule_description text,
                        total_records integer,
                        failed integer,
                        median_30_day_failed double precision,
                        failed_percentage double precision,
                        passed integer,
                        median_30_day_passed double precision,
                        passed_percentage double precision,
                        status text,
                        time_filter text,
                        task_ts timestamp with time zone not null,
                        created_at timestamp with time zone default now() not null,
                        id bigserial not null
                            constraint {table.table_name}_pkey
                                primary key
                    );
                """


class TestMigrationsFrom014to024(MigrationTestCase):
    """
    Test of migration from v 0.1.4. to 0.2.4
    """

    def setUp(self):
        """
        Init a temporary table with some data.
        """
        self.DATA_QUALITY_TABLE_1 = ResultTable(
            DATA_QUALITY_SCHEMA, "example_table", QualityCheck
        )
        self.DATA_QUALITY_TABLE_2 = ResultTable(
            DATA_QUALITY_SCHEMA, "another_table", QualityCheck
        )

        sql = [
            f"DROP SCHEMA IF EXISTS {DATA_QUALITY_SCHEMA} CASCADE;",
            f"CREATE SCHEMA IF NOT EXISTS {DATA_QUALITY_SCHEMA};",
            get_quality_table_creation_script_0_1_4(self.DATA_QUALITY_TABLE_1),
            get_quality_table_creation_script_0_1_4(self.DATA_QUALITY_TABLE_2),
            f"""
            INSERT INTO {self.DATA_QUALITY_TABLE_1.fullname} (attribute, rule_name, rule_description, total_records, failed, median_30_day_failed, failed_percentage, passed, median_30_day_passed, passed_percentage, status, time_filter, task_ts, created_at, id) VALUES ('src', 'not_null', 'True when data is null.', 41136, 0, null, 0, 41136, null, 100, 'valid', null, '2019-11-09 00:00:00.000000', '2019-11-12 12:41:28.391365', 75597);
            INSERT INTO {self.DATA_QUALITY_TABLE_1.fullname} (attribute, rule_name, rule_description, total_records, failed, median_30_day_failed, failed_percentage, passed, median_30_day_passed, passed_percentage, status, time_filter, task_ts, created_at, id) VALUES ('dst', 'not_null', 'True when data is null.', 41136, 0, null, 0, 41136, null, 100, 'valid', null, '2019-11-09 00:00:00.000000', '2019-11-12 12:41:28.391365', 75598);
            INSERT INTO {self.DATA_QUALITY_TABLE_1.fullname} (attribute, rule_name, rule_description, total_records, failed, median_30_day_failed, failed_percentage, passed, median_30_day_passed, passed_percentage, status, time_filter, task_ts, created_at, id) VALUES ('departure_time', 'not_null', 'True when data is null.', 41136, 0, null, 0, 41136, null, 100, 'valid', null, '2019-11-09 00:00:00.000000', '2019-11-12 12:41:28.391365', 75599);
            """,
        ]
        for s in sql:
            self.conn.execute(s)

    def tearDown(self):
        """
        Drop all created tables.
        """
        self.conn.execute(f"DROP schema {DATA_QUALITY_SCHEMA} CASCADE;")
        DQBase.metadata.clear()

    def test_migration_table_is_created(self):
        self.migrate_to("0.2.4")

        migration_table_exists = self.conn.get_records(
            f"""
            SELECT EXISTS (
               SELECT 1
               FROM   information_schema.tables
               WHERE  table_schema = '{self.migration_table.schema_name}'
               AND    table_name = '{self.migration_table.table_name}'
            );            
            """
        )

        assert migration_table_exists

    def test_migration_to_0_2_4(self):
        self.migrate_to("0.2.4")

        rule_type_exists_result = self.conn.get_records(
            f"""
            SELECT EXISTS (
               SELECT 1
               FROM information_schema.columns
               WHERE table_schema='{self.DATA_QUALITY_TABLE_1.schema_name}' and
                     table_name='{self.DATA_QUALITY_TABLE_1.table_name}' and
                     column_name='rule_type'
            );
            """
        )

        assert rule_type_exists_result.first()[0]

        rule_type_exists_result = self.conn.get_records(
            f"""
                    SELECT EXISTS (
                       SELECT 1
                       FROM information_schema.columns
                       WHERE table_schema='{self.DATA_QUALITY_TABLE_2.schema_name}' and
                             table_name='{self.DATA_QUALITY_TABLE_2.table_name}' and
                             column_name='rule_type'
                    );
                    """
        )

        assert rule_type_exists_result.first()[0]

        rule_name_exists_result = self.conn.get_records(
            f"""
                    SELECT EXISTS (
                       SELECT 1
                       FROM information_schema.columns
                       WHERE table_schema='{self.DATA_QUALITY_TABLE_1.schema_name}' and
                             table_name='{self.DATA_QUALITY_TABLE_1.table_name}' and
                             column_name='rule_name'
                    );
                    """
        )

        assert rule_name_exists_result.first()[0]

        rule_name_exists_result = self.conn.get_records(
            f"""
                            SELECT EXISTS (
                               SELECT 1
                               FROM information_schema.columns
                               WHERE table_schema='{self.DATA_QUALITY_TABLE_2.schema_name}' and
                                     table_name='{self.DATA_QUALITY_TABLE_2.table_name}' and
                                     column_name='rule_name'
                            );
                            """
        )

        assert rule_name_exists_result.first()[0]

        rule_type_is_filled_result = self.conn.get_records(
            f"""
                SELECT COUNT(*)
                FROM {self.DATA_QUALITY_TABLE_1.fullname}
                WHERE rule_type='not_null'
            """
        )

        assert rule_type_is_filled_result.first()[0] == 3


class TestMigrations014to024Transaction(MigrationTestCase):
    """
    Test of migration from v 0.1.4. to 0.2.4 transaction
    """

    def setUp(self):
        """
        Init a temporary table with some data.
        """
        self.DATA_QUALITY_TABLE_1 = ResultTable(
            DATA_QUALITY_SCHEMA, "example_table", QualityCheck
        )
        self.DATA_QUALITY_TABLE_2 = ResultTable(
            DATA_QUALITY_SCHEMA, "another_table", QualityCheck
        )

        sql = [
            f"DROP SCHEMA IF EXISTS {DATA_QUALITY_SCHEMA} CASCADE;"
            f"CREATE SCHEMA IF NOT EXISTS {DATA_QUALITY_SCHEMA};",
            get_quality_table_creation_script_0_1_4(self.DATA_QUALITY_TABLE_1),
            get_quality_table_creation_script_0_1_4_invalid(self.DATA_QUALITY_TABLE_2),
            f"""
            INSERT INTO {self.DATA_QUALITY_TABLE_1.fullname} (attribute, rule_name, rule_description, total_records, failed, median_30_day_failed, failed_percentage, passed, median_30_day_passed, passed_percentage, status, time_filter, task_ts, created_at, id) VALUES ('src', 'not_null', 'True when data is null.', 41136, 0, null, 0, 41136, null, 100, 'valid', null, '2019-11-09 00:00:00.000000', '2019-11-12 12:41:28.391365', 75597);
            INSERT INTO {self.DATA_QUALITY_TABLE_1.fullname} (attribute, rule_name, rule_description, total_records, failed, median_30_day_failed, failed_percentage, passed, median_30_day_passed, passed_percentage, status, time_filter, task_ts, created_at, id) VALUES ('dst', 'not_null', 'True when data is null.', 41136, 0, null, 0, 41136, null, 100, 'valid', null, '2019-11-09 00:00:00.000000', '2019-11-12 12:41:28.391365', 75598);
            INSERT INTO {self.DATA_QUALITY_TABLE_1.fullname} (attribute, rule_name, rule_description, total_records, failed, median_30_day_failed, failed_percentage, passed, median_30_day_passed, passed_percentage, status, time_filter, task_ts, created_at, id) VALUES ('departure_time', 'not_null', 'True when data is null.', 41136, 0, null, 0, 41136, null, 100, 'valid', null, '2019-11-09 00:00:00.000000', '2019-11-12 12:41:28.391365', 75599);
            """,
        ]
        for s in sql:
            self.conn.execute(s)

    def tearDown(self):
        """
        Drop all created tables.
        """
        self.conn.execute(f"DROP schema {DATA_QUALITY_SCHEMA} CASCADE;")
        DQBase.metadata.clear()

    def test_migration_to_0_2_4_transaction(self):
        try:
            self.migrate_to("0.2.4")
        except:
            pass

        rule_type_exists_result = self.conn.get_records(
            f"""
            SELECT EXISTS (
               SELECT 1
               FROM information_schema.columns
               WHERE table_schema='{self.DATA_QUALITY_TABLE_1.schema_name}' and
                     table_name='{self.DATA_QUALITY_TABLE_1.table_name}' and
                     column_name='rule_type'
            );
            """
        )

        assert rule_type_exists_result.first()[0] is False

        rule_type_exists_result = self.conn.get_records(
            f"""
                    SELECT EXISTS (
                       SELECT 1
                       FROM information_schema.columns
                       WHERE table_schema='{self.DATA_QUALITY_TABLE_2.schema_name}' and
                             table_name='{self.DATA_QUALITY_TABLE_2.table_name}' and
                             column_name='rule_type'
                    );
                    """
        )

        assert rule_type_exists_result.first()[0] is False

        rule_type_is_filled_result = self.conn.get_records(
            f"""
                SELECT COUNT(*)
                FROM {self.DATA_QUALITY_TABLE_1.fullname}
                WHERE rule_name='not_null'
            """
        )

        assert rule_type_is_filled_result.first()[0] == 3


class TestMigrationTo025(MigrationTestCase):
    def ddl_quality_check_0_2_4(self, table: ResultTable):
        return f"""create table {table.fullname}
                        (
                            attribute text,
                            rule_type text,
                            rule_name text,
                            rule_description text,
                            total_records integer,
                            failed integer,
                            median_30_day_failed double precision,
                            failed_percentage double precision,
                            passed integer,
                            median_30_day_passed double precision,
                            passed_percentage double precision,
                            status text,
                            time_filter text,
                            task_ts timestamp with time zone not null,
                            created_at timestamp with time zone default now() not null,
                            id bigserial not null
                                constraint {table.table_name}_pkey
                                    primary key
                        );
                    """

    def ddl_consistency_check_0_2_4(self, table: ResultTable):
        return f"""create table {table.fullname}
                            (
                                type text,
                                name text,
                                description text,
                                left_table text,
                                right_table text,
                                status text,
                                time_filter text,
                                task_ts timestamp with time zone not null,
                                created_at timestamp with time zone default now() not null,
                                id bigserial not null
                                    constraint {table.table_name}_pkey
                                        primary key
                            );
                        """

    def setUp(self):
        """
        Init a temporary table with some data.
        """
        self.QUALITY_TABLE_1 = ResultTable(DATA_QUALITY_SCHEMA, "table_1", QualityCheck)
        self.CONSISTENCY_TABLE_1 = ResultTable(
            DATA_QUALITY_SCHEMA, "table_2", ConsistencyCheck
        )
        sql = [
            f"DROP SCHEMA IF EXISTS {DATA_QUALITY_SCHEMA} CASCADE;",
            f"CREATE SCHEMA IF NOT EXISTS {DATA_QUALITY_SCHEMA};",
            self.ddl_quality_check_0_2_4(self.QUALITY_TABLE_1),
            self.ddl_consistency_check_0_2_4(self.CONSISTENCY_TABLE_1),
            f"""
                INSERT INTO {self.QUALITY_TABLE_1.fullname}(
                    attribute, rule_name, rule_type, rule_description, total_records,
                    time_filter, task_ts)
                VALUES('a', 'stuff', 'not_null', 'This is the rule.', 10, NULL,
                    \'{FakedDatetime.now().isoformat()}\');
            """
            f"""
                INSERT INTO {self.CONSISTENCY_TABLE_1.fullname}(
                    type, name, description, left_table, right_table, status, time_filter, task_ts
                )
                VALUES(
                    \'{ConsistencyChecker.COUNT}\', 'hello', 'aa', 'tmp.a', 'tmp.b', 'hello',
                    NULL, \'{FakedDatetime.now().isoformat()}\'
                );
            """,
        ]
        for s in sql:
            self.conn.execute(s)

    def tearDown(self):
        """
        Drop all created tables.
        """
        self.conn.execute(f"DROP schema {DATA_QUALITY_SCHEMA} CASCADE;")
        DQBase.metadata.clear()

    def _validate_migration(self, metadata, should_be):
        count = 0
        for name, is_nullable, column_default in metadata:
            count += 1
            assert should_be[name]["is_nullable"] == is_nullable
            if name == "time_filter":
                assert should_be["time_filter"]["column_default"] == column_default
        assert count == len(should_be.keys())

    def _get_metadata(self, cols, table: ResultTable):
        cols_str = ",".join((f"'{c}'" for c in cols))
        return self.conn.execute(
            f"""
            select column_name, is_nullable, column_default
            from information_schema."columns"
            where column_name in ({cols_str})
                and table_name = '{table.table_name}'
                and table_schema = '{table.schema_name}'
        """
        )

    def test_migration_upgrade_to_0_2_5(self):
        """
        NOT NULL set for - attribute, rule_name, rule_type, time_filter.
                         - type, name, left_table, right_table, time_filter
        DEFAULT set for - time_filter (TIME_FILTER_DEFAULT)
        """
        data = self.conn.get_records(
            f"select time_filter from {self.QUALITY_TABLE_1.fullname}"
        )
        assert [d[0] for d in data] == [None]

        data = self.conn.get_records(
            f"select time_filter from {self.CONSISTENCY_TABLE_1.fullname}"
        )
        assert [d[0] for d in data] == [None]

        self.migrate_to_latest()

        # quality check - attribute, rule_name, rule_type, time_filter
        metadata = self._get_metadata(
            ["attribute", "rule_name", "rule_type", "time_filter"], self.QUALITY_TABLE_1
        )
        should_be = {
            "attribute": {"is_nullable": "NO"},
            "rule_name": {"is_nullable": "NO"},
            "rule_type": {"is_nullable": "NO"},
            "time_filter": {
                "is_nullable": "NO",
                "column_default": f"'{TIME_FILTER_DEFAULT}'::text",
            },
        }
        self._validate_migration(metadata, should_be)
        data = self.conn.get_records(
            f"select time_filter from {self.QUALITY_TABLE_1.fullname}"
        )
        assert [d[0] for d in data] == [TIME_FILTER_DEFAULT]

        # consistency check - name, type, left_table, right_table, time_filter
        metadata = self._get_metadata(
            ["type", "name", "left_table", "right_table", "time_filter"],
            self.CONSISTENCY_TABLE_1,
        )
        should_be = {
            "type": {"is_nullable": "NO"},
            "name": {"is_nullable": "NO"},
            "left_table": {"is_nullable": "NO"},
            "right_table": {"is_nullable": "NO"},
            "time_filter": {
                "is_nullable": "NO",
                "column_default": f"'{TIME_FILTER_DEFAULT}'::text",
            },
        }
        self._validate_migration(metadata, should_be)
        data = self.conn.get_records(
            f"select time_filter from {self.CONSISTENCY_TABLE_1.fullname}"
        )
        assert [d[0] for d in data] == [TIME_FILTER_DEFAULT]

    def test_migration_downgrade_to_0_2_4(self):
        self.migrate_to_latest()
        self.migrate_to("0.2.4")

        # quality check
        metadata = self._get_metadata(
            ["attribute", "rule_name", "rule_type", "time_filter"], self.QUALITY_TABLE_1
        )

        should_be = {
            "attribute": {"is_nullable": "YES"},
            "rule_name": {"is_nullable": "YES"},
            "rule_type": {"is_nullable": "YES"},
            "time_filter": {"is_nullable": "YES", "column_default": None},
        }
        self._validate_migration(metadata, should_be)
        data = self.conn.get_records(
            f"select time_filter from {self.QUALITY_TABLE_1.fullname}"
        )
        assert [d[0] for d in data] == [None]

        # consistency check
        metadata = self._get_metadata(
            ["type", "name", "left_table", "right_table", "time_filter"],
            self.CONSISTENCY_TABLE_1,
        )

        should_be = {
            "type": {"is_nullable": "YES"},
            "name": {"is_nullable": "YES"},
            "left_table": {"is_nullable": "YES"},
            "right_table": {"is_nullable": "YES"},
            "time_filter": {"is_nullable": "YES", "column_default": None},
        }
        self._validate_migration(metadata, should_be)
        data = self.conn.get_records(
            f"select time_filter from {self.CONSISTENCY_TABLE_1.fullname}"
        )
        assert [d[0] for d in data] == [None]



import pandas as pd
import pytest

from contessa.executor import refresh_executors, SqlExecutor
from contessa.models import Table
from contessa.rules import (
    SqlRule,
    CustomSqlRule,
    GteRule,
    GtRule,
    NotNullRule,
    NotRule,
    LtRule,
    LteRule,
    EqRule,
)


@pytest.fixture(scope="module")
def df():
    return pd.DataFrame(
        [
            {"src": None, "dst": "abcd", "value": 4},
            {"src": "aa", "dst": "aa", "value": 66},
        ]
    )


@pytest.mark.parametrize(
    "rule, expected",
    [
        (
            GtRule("gt_name", "gt", "value", "value2"),
            [False, False, True, False, False],
        ),  # test another col
        (
            NotNullRule("not_null_name", "not_null", "value"),
            [True, True, True, False, True],
        ),
        (GteRule("gte_name", "gte", "value", 4), [False, True, True, False, True]),
        (NotRule("not_name", "not", "value", 4), [True, False, True, True, False]),
        (LtRule("lt_name", "lt", "value", 4), [True, False, False, False, False]),
        (LteRule("lte_name", "lte", "value", 4), [True, True, False, False, True]),
        (EqRule("eq_name", "eq", "value", 4), [False, True, False, False, True]),
    ],
)
def test_one_column_rule_sql(rule, expected, conn, ctx):
    conn.execute(
        """
            drop table if exists public.tmp_table;

            create table public.tmp_table(
              value int,
              value2 int
            );

            insert into public.tmp_table(value, value2)
            values (1, 2), (4, 5), (5, 3), (NULL, NULL), (4, 11)
        """
    )
    refresh_executors(
        Table(schema_name="public", table_name="tmp_table"), conn, context=ctx
    )

    results = rule.apply(conn)
    expected = pd.Series(expected, name=rule.column)
    assert list(expected) == list(results)


@pytest.mark.parametrize(
    "rule, expected",
    [
        (
            GtRule("gt_name", "gt", "value", 4, condition="conditional is TRUE"),
            [False, False],
        ),
        (
            NotNullRule(
                "not_null_name", "not_null", "value", condition="conditional is TRUE"
            ),
            [True, True],
        ),
        (
            GteRule("gte_name", "gte", "value", 4, condition="conditional is TRUE"),
            [False, True],
        ),
        (
            NotRule("not_name", "not", "value", 4, condition="conditional is TRUE"),
            [True, False],
        ),
        (
            LtRule("lt_name", "lt", "value", 4, condition="conditional is TRUE"),
            [True, False],
        ),
        (
            LteRule("lte_name", "lte", "value", 4, condition="conditional is TRUE"),
            [True, True],
        ),
        (
            EqRule("eq_name", "eq", "value", 4, condition="conditional is TRUE"),
            [False, True],
        ),
        (
            LteRule(
                "lte_name", "lte", "date", "now()", condition="conditional is FALSE"
            ),
            [False, False, True],
        ),
    ],
)
def test_one_column_rule_sql_condition(rule, expected, conn, ctx):
    conn.execute(
        """
            drop table if exists public.tmp_table;

            create table public.tmp_table(
              value int,
              conditional boolean,
              date timestamptz
            );

            insert into public.tmp_table(value, conditional, date)
            values (1, TRUE, NULL), (4, TRUE, NULL), (5, FALSE, NULL), (NULL, FALSE, NULL), (4, FALSE, '2019-10-02T13:30:00+0020')
        """
    )
    refresh_executors(
        Table(schema_name="public", table_name="tmp_table"), conn, context=ctx
    )

    results = rule.apply(conn)
    expected = pd.Series(expected, name=rule.column)
    assert list(expected) == list(results)


@pytest.mark.parametrize(
    "rule, expected",
    [
        (NotRule("not_name", "not", "value1", "value2"), [True, False, False]),
        (LteRule("lte_name", "lte", "value4", "value1"), [True, False, True]),
        (EqRule("eq_name", "eq", "value1", "value3"), [True, True, True]),
        (GtRule("gte_name", "gte", "value2", "value3"), [True, False, False]),
    ],
)
def test_cmp_with_other_col_rules(rule, expected, conn, ctx):
    conn.execute(
        """
        drop table if exists public.tmp_table;

        create table public.tmp_table(
          value1 int,
          value2 int,
          value3 int,
          value4 int
        );

        insert into public.tmp_table(value1, value2, value3, value4)
        values (1, 2, 1, 1), (1, 1, 1, NULL), (1, 1, 1, 1)
    """
    )
    refresh_executors(
        Table(schema_name="public", table_name="tmp_table"), conn, context=ctx
    )

    results = rule.apply(conn)
    expected = pd.Series(expected, name=rule.column)
    assert list(expected) == list(results)


def test_sql_apply(conn, ctx):
    conn.execute(
        """
        drop table if exists public.tmp_table;

        create table public.tmp_table(
          src text,
          dst text
        );

        insert into public.tmp_table(src, dst)
        values ('bts', 'abc'), ('aaa', NULL)
    """
    )
    refresh_executors(
        Table(schema_name="public", table_name="tmp_table"), conn, context=ctx
    )

    sql = """
        select
        src = 'aaa'
        from {{ table_fullname }}
    """
    rule = CustomSqlRule("sql_test_name", "sql_test", "src", sql, "example description")
    results = rule.apply(conn)
    expected = pd.Series([False, True])
    assert list(expected) == list(results)
    conn.execute("""DROP TABLE tmp_table;""")


def test_sql_apply_extra_ctx(conn, ctx):
    ctx["dst_table"] = "public.dst_table"
    conn.execute(
        """
        drop table if exists public.dst_table;

        create table public.dst_table(
          src text,
          dst text,
          created timestamptz
        );

        insert into public.dst_table(src, dst, created)
        values ('bts', 'abc', '2018-09-12T12:00:00'), ('aaa', NULL, '2018-09-12T12:00:00'), ('aaa', NULL, '2019-07-31T12:00:00')
    """
    )
    refresh_executors(Table("public", "dst_table"), conn, context=ctx)

    sql = """
        select
        src = 'aaa'
        from {{ dst_table }}
        where created between timestamptz '{{task_ts}}' and timestamptz '{{task_ts}}' + interval '60 seconds'
    """
    rule = CustomSqlRule(
        "sql_test_name", "sql_test", "col1", sql, "example description"
    )
    results = rule.apply(conn)
    expected = pd.Series([False, True])
    assert list(expected) == list(results)
    conn.execute("""DROP TABLE public.dst_table;""")


def test_new_rule(conn, ctx):
    class CountSqlRule(SqlRule):
        executor_cls = SqlExecutor

        def __init__(self, name, type, count, description=None, **kwargs):
            super().__init__(name, type, description=description, **kwargs)
            self.count = count

        def get_sql_parameters(self):
            context = super().get_sql_parameters()
            context.update({"target_count": self.count})
            return context

        @property
        def sql(self):
            return """
                SELECT COUNT(*) = {{target_count}} FROM {{table_fullname}}
            """

    conn.execute(
        """
        drop table if exists public.tmp_table;
    
        create table public.tmp_table(
          a text,
          b text
        );
    
        insert into public.tmp_table(a, b)
        values ('bts', 'abc'), ('aaa', NULL)
    """
    )

    refresh_executors(
        Table(schema_name="public", table_name="tmp_table"), conn, context=ctx
    )
    rule = CountSqlRule("count_name", "count", 2)
    results = rule.apply(conn)
    expected = pd.Series([True])
    assert list(expected) == list(results)

    rule = CountSqlRule("count_name", "count", 2, condition="a = 'bts'")
    results = rule.apply(conn)
    expected = pd.Series([False])
    assert list(expected) == list(results)
    conn.execute("""DROP TABLE tmp_table;""")



import abc

from contessa.settings import TIME_FILTER_DEFAULT


class Rule(metaclass=abc.ABCMeta):
    """
    Representation of one rule.
    Method apply define how it will be evaluated using the executor that will inject all the
    needed attributes to apply it (df, connector etc.).

    Attributes:
        executor_cls    Executor, indication of which executor class can execute this rule
        description     str, description of the rule

    :param name: str
    :param type: str
    :param description: str
    :param time_filter: str
    :param condition: str
    """

    executor_cls = None
    description = None

    def __init__(
        self, name, type, description, time_filter=TIME_FILTER_DEFAULT, condition=None
    ):
        self.name = name
        self.type = type
        self.time_filter = time_filter
        self.condition = condition
        self.description = description

    @property
    def attribute(self):
        """What attributes(columns) is this rule using."""
        return None

    def apply(self, **kwargs):
        """
        Override this method to check a specific rule.
        """
        raise NotImplementedError

    def __str__(self):
        tf = f"- {self.time_filter}" or ""
        return f"Rule {self.name} of type {self.type} {tf}"



"""
Contessa is data quality framework that can check various rules upon a table. It will store
aggregated results of checks (count of failed/passed etc.) in separate table.

It exposes `DataQualityOperator` that can be plugged in any airflow DAG and will do data quality
check on load in temporary table.
"""

__version__ = "0.2.5"

# Start ignoring PyUnusedCodeBear
from .consistency_checker import ConsistencyChecker
from .runner import ContessaRunner
from .rules import EQ, GT, GTE, LT, LTE, NOT, NOT_COLUMN, NOT_NULL, SQL

# Stop ignoring



import abc
from datetime import datetime, timedelta
import logging
from typing import Dict

import pandas as pd

from contessa.db import Connector
from contessa.models import Table
from contessa.utils import compose_where_time_filter


class Executor(metaclass=abc.ABCMeta):
    """
    Class that execute a rule and gives a proper kwargs to the `Rule.apply()` method.
    It also holds the `raw_df` and `filtered_df`.
    """

    date_columns = ("created_at", "updated_at", "confirmed_at")

    def __init__(self, check_table: Table, conn: Connector, context: Dict):
        self.conn = conn
        self.check_table = check_table
        self.context = context
        self._raw_df = None

    def matched_cols(self, cols):
        for col in cols:
            if col in self.date_columns:
                yield col

    @property
    def raw_df(self):
        """
        Cache raw df of temporary table we are doing quality check on.
        Casts datetime fields to python datetime.
        :return pd.Dataframe
        """
        if self._raw_df is not None:
            return self._raw_df
        self._raw_df = self.conn.get_pandas_df(
            f"select * from {self.check_table.fullname}"
        )

        # cast datetime cols to python datetime
        # pandas issue...
        for col in self.matched_cols(self._raw_df.columns):
            try:
                self._raw_df[col] = self._raw_df[col].apply(
                    lambda d: pd.to_datetime(str(d)).replace(tzinfo=None)
                )
            except:
                logging.warning(
                    f"Wrong date in `{col}` of {self.check_table.fullname}. Probably all Nones."
                )
        return self._raw_df

    def compose_where_time_filter(self, rule):
        """
        Composes WHERE statement, which filters records by time_filter`.
        Rule attribute `time_filter` filters
        only data that were updated/created/confirmed in last 30 days.
        :return: str, WHERE `time_filter` filter statement
        """
        return compose_where_time_filter(rule.time_filter, self.context["task_ts"])

    def compose_where_condition(self, rule):
        """
        Composes WHERE statement, which filters records by user-provided condition with Rule attribute `condition`.
        Very simple for now, ready for possible extensions later.
        :return: str, WHERE `condition` filter statement
        """
        condition = rule.condition
        if condition:
            return condition
        else:
            return ""

    def filter_df(self, rule):
        """
        Filter the `raw_df` to be give it to apply method. Rule attribute `time_filter` filters
        only data that were updated/created/confirmed in last 30 days.
        :return: pd.Dataframe
        """
        md = rule.time_filter
        if md:
            past = datetime.now() - timedelta(days=30)
            selector = self.raw_df[md] >= past
            filtered_df = self.raw_df[selector]
        else:
            filtered_df = self.raw_df
        return filtered_df

    def execute(self, rule):
        """
        Main entrypoint for Executor class. Composing kwargs specific for the executor
        (e.g. PandasExecutor) and get the results from rule.
        :param rule: Rule
        :return: pd.Series
        """
        kwargs = self.compose_kwargs(rule)
        results = rule.apply(**kwargs)
        return results

    @abc.abstractmethod
    def compose_kwargs(self, rule):
        raise NotImplementedError


class PandasExecutor(Executor):
    def compose_kwargs(self, rule):
        return {"df": self.filter_df(rule)}


class SqlExecutor(Executor):
    def compose_kwargs(self, rule):
        return {"conn": self.conn}


# singleton of executors
executors = None


def refresh_executors(check_table: Table, conn: Connector, context: Dict):
    """
    Use this to re-init the executor classes that are used to execute rules. To have right
    data from new table in Executor class.

    NOTE: This would fail if it 2 quality checks would run in 2 threads or async manner.
    Should exists a pool of executors that are mapped using schema and table_names.

    """
    global executors
    executors = {
        PandasExecutor: PandasExecutor(check_table, conn, context),
        SqlExecutor: SqlExecutor(check_table, conn, context),
    }
    logging.info("Successfully inited PandasExecutor and SqlExecutor.")


def get_executor(rule):
    """
    Return instance of Executor for a specific Rule.
    :param rule: Rule
    :return: Executor
    """
    return executors[rule.executor_cls]



import itertools
import logging


class RuleNormalizer:
    """
    Class that split list values to normalized rules that checks one rule, one col,
    one time_filter with provided where condition.

    Example:
        {'type': 'not_null', 'columns': ['a', 'b', 'c'], 'separate_time_filters': ['c', 'u'], 'condition': 'd is TRUE'}

        will be transformed to

        {'type': 'not_null', 'column': 'a', 'time_filter': 'c', 'condition': 'd is TRUE'}
        {'type': 'not_null', 'column': 'b', 'time_filter': 'c', 'condition': 'd is TRUE'}
        {'type': 'not_null', 'column': 'c', 'time_filter': 'c', 'condition': 'd is TRUE'}
        {'type': 'not_null', 'column': 'a', 'time_filter': 'u', 'condition': 'd is TRUE'}
        {'type': 'not_null', 'column': 'b', 'time_filter': 'u', 'condition': 'd is TRUE'}
        {'type': 'not_null', 'column': 'c', 'time_filter': 'u', 'condition': 'd is TRUE'}

    This class should define what we can support as an input to be able to make Rule classes
    from it. Currently it supports:

        - multiple columns
        - multiple separate_time_filters

    """

    @classmethod
    def normalize(cls, rules_def):
        normalized = []
        for rule_def in rules_def:
            # if it is normalized, we skip it
            if not cls._should_normalize(rule_def):
                normalized.append(rule_def)
                continue
            # otherwise we split all the permutations of lists into separate rules
            permutations = cls._get_permutations(rule_def)
            new_rules = cls._split_permutations(permutations, rule_def)
            normalized.extend(new_rules)
        return normalized

    @staticmethod
    def _should_normalize(rule_def):
        if "columns" in rule_def:
            return True
        elif "separate_time_filters" in rule_def:
            if len(rule_def["separate_time_filters"]) <= 1:
                raise ValueError("Please use `time_filter` for one column.")
            return True
        return False

    @staticmethod
    def _get_permutations(rule_def):
        cols = rule_def.get("columns") or [rule_def.get("column")] or [None]
        time_filters = (
            rule_def.get("separate_time_filters")
            or [rule_def.get("time_filter")]
            or [None]
        )
        permutations = itertools.product(cols, time_filters)
        return permutations

    @staticmethod
    def _split_permutations(permutations, rule_def):
        new_rules = []
        for perm in permutations:
            # time_filter has to be one of list, str or None
            if isinstance(perm[1], dict):
                time_filter = [perm[1]]
            else:
                time_filter = perm[1]
            tmp = rule_def.copy()
            tmp["column"] = perm[0]
            tmp.pop("columns", None)
            tmp["time_filter"] = time_filter
            tmp.pop("separate_time_filters", None)
            new_rules.append(tmp)
        return new_rules



import pandas as pd

from contessa.db import Connector
from contessa.models import DQBase
from test.conftest import FakedDatetime
from test.integration.conftest import TEST_DB_URI
import unittest
from unittest import mock

from contessa import ContessaRunner


class TestDataQualityOperator(unittest.TestCase):
    def setUp(self):
        """
        Init a temporary table with some data.
        """
        self.table_name = "booking_all_v2"
        self.ts_nodash = (
            FakedDatetime.now().isoformat().replace("-", "").replace(":", "")
        )
        self.tmp_table_name = f"{self.table_name}_{self.ts_nodash}"
        self.now = FakedDatetime.now()

        sql = [
            "DROP SCHEMA if exists tmp CASCADE;",
            "DROP SCHEMA if exists data_quality CASCADE;",
            "CREATE SCHEMA IF NOT EXISTS tmp;",
            "CREATE SCHEMA IF NOT EXISTS data_quality;",
            "CREATE SCHEMA IF NOT EXISTS hello;",
            f"""
                    CREATE TABLE IF NOT EXISTS tmp.{self.table_name}(
                      id SERIAL PRIMARY KEY,
                      src text,
                      dst text,
                      price int,
                      turnover_after_refunds double precision,
                      initial_price double precision,
                      created_at timestamptz
                    )
                """,
            f"""
                    CREATE TABLE IF NOT EXISTS tmp.{self.tmp_table_name}(
                      id SERIAL PRIMARY KEY,
                      src text,
                      dst text,
                      price int,
                      turnover_after_refunds double precision,
                      initial_price double precision,
                      created_at timestamptz
                    )
                """,
            f"""
                    INSERT INTO tmp.{self.tmp_table_name}
                        (src, dst, price, turnover_after_refunds, initial_price, created_at)
                    VALUES
                        ('BTS', NULL, 1, 100, 11, '2018-09-12T11:50:00'),
                        -- this is older than 30 days.
                        -- not in stats when time_filter = `created_at`
                        (NULL, 'PEK', 33, 1.1, 13, '2018-01-12T15:50:00'),
                        ('VIE', 'JFK', 4, 5.5, 23.4, '2018-09-11T11:50:00'),
                        ('VIE', 'VIE', 4, 0.0, 0.0, '2018-09-11T11:50:00')
                """,
            f"""
                INSERT INTO tmp.{self.table_name}
                    (src, dst, price, turnover_after_refunds, initial_price, created_at)
                VALUES
                    ('BTS', NULL, 1, 100, 11, '2018-09-12T13:00:00'),
                    -- this is older than 30 days.
                    -- not in stats when time_filter = `created_at`
                    (NULL, 'PEK', 33, 1.1, 13, '2018-01-12T13:00:00'),
                    ('VIE', 'JFK', 4, 5.5, 23.4, '2018-09-11T13:00:00'),
                    ('VIE', 'VIE', 4, 0.0, 0.0, '2018-09-11T13:00:00')
            """,
        ]
        self.conn = Connector(TEST_DB_URI)
        for s in sql:
            self.conn.execute(s)

        self.contessa_runner = ContessaRunner(TEST_DB_URI)

    def tearDown(self):
        """
        Drop all created tables.
        """
        self.conn.execute(f"DROP schema tmp CASCADE;")
        self.conn.execute(f"DROP schema data_quality CASCADE;")
        self.conn.execute(f"DROP schema hello CASCADE;")
        DQBase.metadata.clear()

    @mock.patch("contessa.executor.datetime", FakedDatetime)
    def test_execute_tmp(self):
        sql = """
            SELECT
              CASE WHEN src = 'BTS' and dst is null THEN false ELSE true END as res
            from {{ table_fullname }}
        """
        rules = [
            {
                "name": "not_null_name",
                "type": "not_null",
                "column": "dst",
                "time_filter": "created_at",
            },
            {
                "name": "gt_name",
                "type": "gt",
                "column": "price",
                "value": 10,
                "time_filter": "created_at",
            },
            {
                "name": "sql_name",
                "type": "sql",
                "sql": sql,
                "column": "src_dst",
                "description": "test sql rule",
            },
            {"name": "not_name", "type": "not", "column": "src", "value": "dst"},
        ]
        self.contessa_runner.run(
            check_table={"schema_name": "tmp", "table_name": self.tmp_table_name},
            result_table={"schema_name": "data_quality", "table_name": self.table_name},
            raw_rules=rules,
            context={"task_ts": self.now},
        )

        rows = self.conn.get_pandas_df(
            f"""
            SELECT * from data_quality.quality_check_{self.table_name}
            order by created_at
        """
        )
        self.assertEqual(rows.shape[0], 4)

        notnull_rule = rows.loc[0]
        self.assertEqual(notnull_rule["failed"], 1)
        self.assertEqual(notnull_rule["passed"], 2)
        self.assertEqual(notnull_rule["attribute"], "dst")
        self.assertEqual(notnull_rule["task_ts"].timestamp(), self.now.timestamp())

        gt_rule = rows.loc[1]
        self.assertEqual(gt_rule["failed"], 3)
        self.assertEqual(gt_rule["passed"], 0)
        self.assertEqual(gt_rule["attribute"], "price")

        sql_rule = rows.loc[2]
        self.assertEqual(sql_rule["failed"], 1)
        self.assertEqual(sql_rule["passed"], 3)
        self.assertEqual(sql_rule["attribute"], "src_dst")

        not_column_rule = rows.loc[3]
        self.assertEqual(not_column_rule["failed"], 1)
        self.assertEqual(not_column_rule["passed"], 3)
        self.assertEqual(not_column_rule["attribute"], "src")

    @mock.patch("contessa.executor.datetime", FakedDatetime)
    def test_execute_dst(self):
        sql = """
            SELECT
              CASE WHEN src = 'BTS' AND dst is null THEN false ELSE true END as res
            FROM {{ table_fullname }}
            WHERE created_at BETWEEN
                timestamptz '{{task_ts}}' - INTERVAL '1 hour'
                AND
                timestamptz '{{task_ts}}' + INTERVAL '1 hour'
        """
        rules = [
            {
                "name": "not_null_name",
                "type": "not_null",
                "column": "dst",
                "time_filter": "created_at",
            },
            {
                "name": "sql_name",
                "type": "sql",
                "sql": sql,
                "column": "src_dst",
                "description": "test sql rule",
            },
        ]
        self.contessa_runner.run(
            check_table={"schema_name": "tmp", "table_name": self.tmp_table_name},
            result_table={"schema_name": "data_quality", "table_name": self.table_name},
            raw_rules=rules,
            context={"task_ts": self.now},
        )

        rows = self.conn.get_pandas_df(
            f"""
            SELECT * from data_quality.quality_check_{self.table_name}
            order by created_at
        """
        )
        self.assertEqual(rows.shape[0], 2)

        notnull_rule = rows.loc[0]
        self.assertEqual(notnull_rule["failed"], 1)
        self.assertEqual(notnull_rule["passed"], 2)
        self.assertEqual(notnull_rule["attribute"], "dst")
        self.assertEqual(notnull_rule["task_ts"].timestamp(), self.now.timestamp())

        sql_rule = rows.loc[1]
        self.assertEqual(sql_rule["failed"], 1)
        self.assertEqual(sql_rule["passed"], 0)
        self.assertEqual(sql_rule["attribute"], "src_dst")

    def test_different_schema(self):
        rules = [
            {
                "name": "not_nul_name",
                "type": "not_null",
                "column": "dst",
                "time_filter": "created_at",
            }
        ]
        self.contessa_runner.run(
            check_table={"schema_name": "tmp", "table_name": self.tmp_table_name},
            result_table={"schema_name": "hello", "table_name": "abcde"},
            raw_rules=rules,
            context={"task_ts": self.now},
        )
        rows = self.conn.get_pandas_df(
            f"""
                SELECT 1 from hello.quality_check_abcde
            """
        )
        self.assertEqual(rows.shape[0], 1)



from contessa.migration import MigrationsResolver

SQLALCHEMY_URL = "postgresql://postgres:postgres@postgres:5432/test_db"


def test_get_fallback_version_present_in_map():
    versions_migrations = {"0.1.2": "w5rtyuret457", "0.1.3": "dfgdfg5b0ee5"}

    m = MigrationsResolver(versions_migrations, "0.1.3", SQLALCHEMY_URL, "schema")
    fallback_version = m.get_fallback_version()

    assert fallback_version == "0.1.3"


def test_get_fallback_version_less_than_first():
    versions_migrations = {"0.1.2": "w5rtyuret457", "0.1.3": "dfgdfg5b0ee5"}

    m = MigrationsResolver(versions_migrations, "0.1.1", SQLALCHEMY_URL, "schema")
    fallback_version = m.get_fallback_version()

    assert fallback_version == "0.1.2"


def test_get_fallback_version_greather_than_last():
    versions_migrations = {"0.1.2": "w5rtyuret457", "0.1.3": "dfgdfg5b0ee5"}

    m = MigrationsResolver(versions_migrations, "0.1.8", SQLALCHEMY_URL, "schema")
    fallback_version = m.get_fallback_version()

    assert fallback_version == "0.1.3"


def test_get_fallback_version_is_betwen():
    versions_migrations = {
        "0.1.2": "w5rtyuret457",
        "0.1.6": "dfgdfg5b0ee5",
        "0.1.7": "54f8985b0ee5",
        "0.1.9": "480e6618700d",
        "0.2.6": "3w4er8y50yyd",
        "0.2.8": "034hfa8943hr",
    }

    m = MigrationsResolver(versions_migrations, "0.1.8", SQLALCHEMY_URL, "schema")
    fallback_version = m.get_fallback_version()

    assert fallback_version == "0.1.7"

    m = MigrationsResolver(versions_migrations, "0.2.0", SQLALCHEMY_URL, "schema")
    fallback_version = m.get_fallback_version()

    assert fallback_version == "0.1.9"



import logging
from typing import List, Dict, Optional

from datetime import datetime

from contessa.base_rules import Rule
from contessa.db import Connector
from contessa.executor import get_executor, refresh_executors
from contessa.models import create_default_check_class, Table, ResultTable, QualityCheck
from contessa.normalizer import RuleNormalizer
from contessa.rules import get_rule_cls


class ContessaRunner:
    model_cls = QualityCheck

    def __init__(self, conn_uri_or_engine, special_qc_map=None):
        self.conn_uri_or_engine = conn_uri_or_engine
        self.conn = Connector(conn_uri_or_engine)

        # todo - allow cfg
        self.special_qc_map = special_qc_map or {}

    def run(
        self,
        raw_rules: List[Dict[str, str]],
        check_table: Dict,
        result_table: Dict,  # todo - docs for quality name, maybe defaults..
        context: Optional[Dict] = None,
    ):
        check_table = Table(**check_table)
        result_table = ResultTable(**result_table, model_cls=self.model_cls)
        context = self.get_context(check_table, context)

        normalized_rules = self.normalize_rules(raw_rules)
        refresh_executors(check_table, self.conn, context)
        quality_check_class = self.get_quality_check_class(result_table)
        self.conn.ensure_table(quality_check_class.__table__)

        rules = self.build_rules(normalized_rules)
        objs = self.do_quality_checks(quality_check_class, rules, context)

        self.conn.upsert(objs)

    @staticmethod
    def get_context(check_table: Table, context: Optional[Dict] = None) -> Dict:
        """
        Construct context to pass to executors. User context overrides defaults.
        """
        ctx_defaults = {
            "table_fullname": check_table.fullname,
            "task_ts": datetime.now(),  # todo - is now() ok ?
        }
        ctx_defaults.update(context)
        return ctx_defaults

    def normalize_rules(self, raw_rules):
        return RuleNormalizer.normalize(raw_rules)

    def do_quality_checks(self, dq_cls, rules: List[Rule], context: Dict = None):
        """
        Run quality check for all rules. Use `qc_cls` to construct objects that will be inserted
        afterwards.
        """
        ret = []
        for rule in rules:
            obj = self.apply_rule(context, dq_cls, rule)
            ret.append(obj)
        return ret

    def apply_rule(self, context, dq_cls, rule):
        e = get_executor(rule)
        logging.info(f"Executing rule `{rule}`.")
        results = e.execute(rule)
        obj = dq_cls()
        obj.init_row(rule, results, self.conn, context)
        return obj

    @staticmethod
    def build_rules(normalized_rules):
        """
        Construct rules classes from user definition that are dicts.
        Raises if there are bad arguments for a certain rule.
        :return: list of Rule objects
        """
        ret = []
        for rule_def in normalized_rules:
            rule_cls = ContessaRunner.pick_rule_cls(rule_def)
            try:
                r = rule_cls(**rule_def)
            except Exception as e:
                logging.error(f"For rule `{rule_cls.__name__}`. {e.args[0]}")
                raise
            else:
                ret.append(r)
        return ret

    @staticmethod
    def pick_rule_cls(rule_def):
        """
        Get rule class based on its type that was input by user.
        :param rule_def: dict
        :return: Rule class
        """
        return get_rule_cls(rule_def["type"])

    def get_quality_check_class(self, result_table: ResultTable):
        """
        QualityCheck can be different, e.g. `special_table` has specific quality_check.
        Or kind of generic one that computes number of passed/failed objects etc.
        So determine if is special or not and return the class.
        :return: QualityCheck cls
        """
        special_checks = self.special_qc_map.keys()
        if result_table.fullname in special_checks:
            quality_check_class = self.special_qc_map[result_table.fullname]
            logging.info(
                f"Using {quality_check_class.__name__} as quality check class."
            )
        else:
            quality_check_class = create_default_check_class(result_table)
            logging.info("Using default QualityCheck class.")
        return quality_check_class



from datetime import datetime

import pandas as pd
import pytest

from contessa.rules import NotNullRule


class FakedDatetime(datetime):
    @classmethod
    def now(cls, **kwargs):
        return cls(2018, 9, 12, 12, 0, 0)

    @classmethod
    def today(cls):
        return cls(2018, 9, 12, 12, 0, 0)


@pytest.fixture(scope="session")
def rule():
    return NotNullRule("not_null_name", "not_null", "src")


@pytest.fixture(scope="session")
def results():
    return pd.Series([True, True, False, False, True], name="src")


@pytest.fixture(scope="session")
def ctx():
    return {"task_ts": FakedDatetime.now(), "table_fullname": "public.tmp_table"}



from datetime import datetime, timedelta

import pandas as pd

from contessa.executor import PandasExecutor, SqlExecutor
from contessa.models import Table
from contessa.rules import NotNullRule


def test_compose_kwargs_sql_executor(dummy_contessa, ctx):
    t = Table(**{"schema_name": "tmp", "table_name": "hello_world"})
    e = SqlExecutor(t, dummy_contessa.conn, ctx)
    rule = NotNullRule("not_null_name", "not_null", "src", time_filter="created_at")
    kwargs = e.compose_kwargs(rule)
    expected = {"conn": dummy_contessa.conn}
    assert kwargs == expected


def test_compose_kwargs_sql_executor_time_filter(dummy_contessa, ctx):
    t = Table(**{"schema_name": "tmp", "table_name": "hello_world"})
    e = SqlExecutor(t, dummy_contessa.conn, ctx)

    rule = NotNullRule("not_null_name", "not_null", "src", time_filter="created_at")
    time_filter = e.compose_where_time_filter(rule)
    computed_datetime = (ctx["task_ts"] - timedelta(days=30)).strftime(
        "%Y-%m-%d %H:%M:%S"
    )
    expected = f"(created_at >= '{computed_datetime} UTC'::timestamptz AND created_at < '{ctx['task_ts']} UTC'::timestamptz)"
    assert time_filter == expected, "time_filter is string"

    rule = NotNullRule(
        "not_null_name", "not_null", "src", time_filter=[{"column": "created_at"}]
    )
    time_filter = e.compose_where_time_filter(rule)
    computed_datetime = (ctx["task_ts"] - timedelta(days=30)).strftime(
        "%Y-%m-%d %H:%M:%S"
    )
    expected = f"(created_at >= '{computed_datetime} UTC'::timestamptz AND created_at < '{ctx['task_ts']} UTC'::timestamptz)"
    assert time_filter == expected, "time_filter has only column"

    rule = NotNullRule(
        "not_null_name",
        "not_null",
        "src",
        time_filter=[
            {"column": "created_at", "days": 10},
            {"column": "updated_at", "days": 1},
        ],
    )
    time_filter = e.compose_where_time_filter(rule)
    computed_created = (ctx["task_ts"] - timedelta(days=10)).strftime(
        "%Y-%m-%d %H:%M:%S"
    )
    computed_updated = (ctx["task_ts"] - timedelta(days=1)).strftime(
        "%Y-%m-%d %H:%M:%S"
    )
    expected = (
        f"(created_at >= '{computed_created} UTC'::timestamptz AND created_at < '{ctx['task_ts']} UTC'::timestamptz) OR "
        f"(updated_at >= '{computed_updated} UTC'::timestamptz AND updated_at < '{ctx['task_ts']} UTC'::timestamptz)"
    )
    assert time_filter == expected, "time_filter has 2 members"


def test_compose_kwargs_pd_executor(dummy_contessa, ctx):
    t = Table(**{"schema_name": "tmp", "table_name": "hello_world"})
    e = PandasExecutor(t, dummy_contessa.conn, ctx)
    rule = NotNullRule("not_null_name", "not_null", "src", time_filter="created_at")
    df = pd.DataFrame([{"created_at": datetime(2017, 10, 10)}])
    e.conn.get_pandas_df = lambda x: df
    kwargs = e.compose_kwargs(rule)
    expected = {"df": df}
    assert kwargs.keys() == expected.keys()



import logging
from typing import Dict, Optional, List

import jinja2
import sqlalchemy
from datetime import datetime

from contessa.db import Connector
from contessa.models import (
    create_default_check_class,
    Table,
    ResultTable,
    ConsistencyCheck,
)
from contessa.settings import TIME_FILTER_DEFAULT
from contessa.utils import compose_where_time_filter


class ConsistencyChecker:
    """
    Checks consistency of the sync between two tables.
    """

    model_cls = ConsistencyCheck

    COUNT = "count"
    DIFF = "difference"

    def __init__(self, left_conn_uri_or_engine, right_conn_uri_or_engine=None):
        self.left_conn_uri_or_engine = left_conn_uri_or_engine
        self.left_conn = Connector(left_conn_uri_or_engine)
        if right_conn_uri_or_engine is None:
            self.right_conn_uri_or_engine = self.left_conn_uri_or_engine
            self.right_conn = self.left_conn
        else:
            self.right_conn_uri_or_engine = right_conn_uri_or_engine
            self.right_conn = Connector(right_conn_uri_or_engine)

    def run(
        self,
        method: str,
        left_check_table: Dict,
        right_check_table: Dict,
        result_table: Dict,
        columns: Optional[List[str]] = None,
        time_filter: str = TIME_FILTER_DEFAULT,
        left_custom_sql: str = None,
        right_custom_sql: str = None,
        context: Optional[Dict] = None,
    ):
        if left_custom_sql and right_custom_sql:
            if columns or time_filter:
                raise ValueError(
                    "When using custom sqls you cannot change 'columns' or 'time_filter' attribute"
                )

        left_check_table = Table(**left_check_table)
        right_check_table = Table(**right_check_table)
        result_table = ResultTable(**result_table, model_cls=self.model_cls)
        context = self.get_context(left_check_table, right_check_table, context)

        result = self.do_consistency_check(
            method,
            columns,
            time_filter,
            left_check_table,
            right_check_table,
            left_custom_sql,
            right_custom_sql,
            context,
        )

        quality_check_class = create_default_check_class(result_table)
        self.right_conn.ensure_table(quality_check_class.__table__)
        self.upsert(quality_check_class, result)

    @staticmethod
    def get_context(
        left_check_table: Table,
        right_check_table: Table,
        context: Optional[Dict] = None,
    ) -> Dict:
        """
        Construct context to pass to executors. User context overrides defaults.
        """
        ctx_defaults = {
            "left_table_fullname": left_check_table.fullname,
            "right_table_fullname": right_check_table.fullname,
            "task_ts": datetime.now(),
        }
        ctx_defaults.update(context)
        return ctx_defaults

    def do_consistency_check(
        self,
        method: str,
        columns: Optional[List[str]],
        time_filter: str,
        left_check_table: Table,
        right_check_table: Table,
        left_sql: str = None,
        right_sql: str = None,
        context: Dict = None,
    ):
        """
        Run quality check for all rules. Use `qc_cls` to construct objects that will be inserted
        afterwards.
        """
        if not left_sql or not right_sql:
            if method == self.COUNT:
                if columns:
                    column = f"count({', '.join(columns)})"
                else:
                    column = "count(*)"
            elif method == self.DIFF:
                if columns:
                    column = ", ".join(columns)
                else:
                    # List the columns explicitly in case column order of compared tables is not the same.
                    column = ", ".join(
                        sorted(
                            self.right_conn.get_column_names(right_check_table.fullname)
                        )
                    )
            else:
                raise NotImplementedError(f"Method {method} not implemented")

        if not left_sql:
            left_sql = self.construct_default_query(
                left_check_table.fullname, column, time_filter, context
            )
        left_result = self.run_query(self.left_conn, left_sql, context)
        if not right_sql:
            right_sql = self.construct_default_query(
                right_check_table.fullname, column, time_filter, context
            )
        right_result = self.run_query(self.right_conn, right_sql, context)

        return {
            "check": {"type": method, "description": "", "name": "consistency",},
            "status": "valid"
            if self.compare_results(left_result, right_result)
            else "invalid",
            "left_table_name": left_check_table.fullname,
            "right_table_name": right_check_table.fullname,
            "time_filter": time_filter,
            "context": context,
        }

    def compare_results(self, left_result, right_result):
        return set(left_result) == set(right_result)

    def construct_default_query(self, table_name, column, time_filter, context):
        time_filter = compose_where_time_filter(time_filter, context["task_ts"])
        query = f"""
            SELECT {column}
            FROM {table_name}
            {f'WHERE {time_filter}' if time_filter else ''}
        """
        return query

    def render_sql(self, sql, context):
        """
        Replace some parameters in query.
        :return str, formatted sql
        """
        t = jinja2.Template(sql)
        rendered = t.render(**context)
        return rendered

    def run_query(self, conn: Connector, query: str, context):
        query = self.render_sql(query, context)
        logging.info(query)
        result = [r._row for r in conn.get_records(query)]
        return result

    def upsert(self, dc_cls, result):
        obj = dc_cls()
        obj.init_row(**result)
        self.right_conn.upsert(
            [obj,]
        )



from datetime import datetime

import pandas as pd
import pytest

from contessa.executor import Executor
from contessa.models import Table
from contessa.rules import NotNullRule
from test.conftest import FakedDatetime


@pytest.fixture(autouse=True)
def db(conn):
    conn.execute(
        """
        create table tmp.hello_world(
          src text,
          dst text,
          value double PRECISION,
          created_at timestamptz default NULL
        );

        insert into tmp.hello_world(src, dst, value, created_at)
        values ('a', 'b', 3, '2018-09-12T13:00:00'), ('aa', 'bb', 55, NULL);
    """
    )
    yield

    conn.execute("DROP table tmp.hello_world")


@pytest.fixture
def e(conn, ctx):
    class ConcreteExecutor(Executor):
        def compose_kwargs(self, rule):
            return {}

    return ConcreteExecutor(Table("tmp", "hello_world"), conn, ctx)


def test_executor_raw_df(e):
    df = e.raw_df
    expected = pd.DataFrame(
        [("a", "b", 3.0, datetime(2018, 9, 12, 13)), ("aa", "bb", 55.0, None)],
        columns=["src", "dst", "value", "created_at"],
    )
    assert df.equals(expected)


def test_executor_filter_df(e, monkeypatch):
    rule = NotNullRule("not_null_name", "not_null", "src", time_filter="created_at")
    monkeypatch.setattr("contessa.executor.datetime", FakedDatetime)
    df = e.filter_df(rule)
    expected = pd.DataFrame(
        [("a", "b", 3.0, datetime(2018, 9, 12, 13))],
        columns=["src", "dst", "value", "created_at"],
    )
    assert df.equals(expected)



"""change type to name

Revision ID: 480e6618700d
Revises: 54f8985b0ee5
Create Date: 2019-11-04 14:12:21.483875

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy import create_engine, inspect

# revision identifiers, used by Alembic.
from contessa.models import QualityCheck, ConsistencyCheck

revision = "480e6618700d"
down_revision = "54f8985b0ee5"
branch_labels = None
depends_on = None

config = None


def get_config():
    global config
    if config:
        return config

    from alembic import context

    config = context.config

    return config


def get(name):
    return get_config().get_main_option(name)


def get_quality_tables(table_prefix):
    url = get("sqlalchemy.url")
    schema = get("schema")

    engine = create_engine(url)
    inspector = inspect(engine)

    all_tables = inspector.get_table_names(schema=schema)
    quality_tables = [x for x in all_tables if x.startswith(table_prefix)]

    return quality_tables


def upgrade():
    schema = get("schema")

    print("Migration Quality Check")
    for table in get_quality_tables(QualityCheck._table_prefix):
        print(f"Migrate table {table}")
        op.alter_column(
            table,
            "rule_name",
            nullable=False,
            new_column_name="rule_type",
            schema=schema,
        )

        op.add_column(table, sa.Column("rule_name", sa.TEXT), schema=schema)


def downgrade():
    schema = get("schema")

    print("Migration Quality Check")
    for table in get_quality_tables(QualityCheck._table_prefix):
        print(f"Migrate table {table}")

        op.drop_column(table, "rule_name", schema=schema)
        op.alter_column(
            table,
            "rule_type",
            nullable=False,
            new_column_name="rule_name",
            schema=schema,
        )



from typing import Union, List
import logging

from sqlalchemy import create_engine, Table, UniqueConstraint
from sqlalchemy.dialects.postgresql import insert
from sqlalchemy.engine.base import Engine
import pandas.io.sql as pdsql
from sqlalchemy.orm import sessionmaker


class Connector:
    """
    Wrapping sqlachemy engine. Holds some useful methods.
    """

    def __init__(self, conn_uri_or_engine: Union[str, Engine]):
        if isinstance(conn_uri_or_engine, str):
            self.engine = create_engine(conn_uri_or_engine)
        elif isinstance(conn_uri_or_engine, Engine):
            self.engine = conn_uri_or_engine
        else:
            cls_name = self.__class__.__name__
            raise ValueError(
                f"You can only pass conn str or sqlalchemy `Engine` to `{cls_name}`."
            )
        self.Session = sessionmaker(bind=self.engine)

    def make_session(self):
        return self.Session()

    def get_records(self, sql, params=None):
        """
        Just proxy with better name if used.
        """
        return self.execute(sql, params)

    def execute(self, sql: [List, str], params=None):
        """
        Execute sql, if there are some results, return them.
        """
        params = params or {}
        with self.engine.connect() as conn:
            rs = conn.execute(sql, **params)
        return rs

    def get_pandas_df(self, sql):
        return pdsql.read_sql(sql, con=self.engine)

    def ensure_table(self, table: Table):
        """
        Create table for given table class if it doesn't exists.
        """
        table.create(bind=self.engine, checkfirst=True)
        logging.info(f"Created table {table.name}.")

    @staticmethod
    def model2dict(obj):
        """
        Model instance dict contains all the cols, but also internal _sa_instance_state.
        """
        a = obj.__dict__.copy()
        a.pop("_sa_instance_state", None)
        return a

    def upsert(self, objs):
        """
        Insert on conflict do update.
        """
        logging.info(f"Upserting {len(objs)} results.")

        data = []
        for o in objs:
            data.append(self.model2dict(o))

        table = objs[0].__table__
        stmt = insert(table).values(data)
        conflicting_cols = get_unique_constraint_names(table)
        excluded_set = {k: getattr(stmt.excluded, k) for k in data[0].keys()}

        on_update_stmt = stmt.on_conflict_do_update(
            index_elements=conflicting_cols, set_=excluded_set
        )

        session = self.make_session()
        try:
            session.execute(on_update_stmt)
            session.commit()
        except:
            session.rollback()
            raise
        finally:
            session.close()

    def get_column_names(self, table_full_name: str) -> List:
        schema_query = f"""
                SELECT
                    column_name
                FROM information_schema.columns
                WHERE concat(table_schema, '.', table_name) = '{table_full_name}'
                ORDER BY ordinal_position
            """

        return [col[0] for col in self.get_records(schema_query)]


def get_unique_constraint_names(table):
    """
    Doesn't make sense if there are multiple unique constraints, as nothing indicate which one
    to pick. If there is only 1, return names of the columns.
    """
    unique_constraint = [
        u for u in table.constraints if isinstance(u, UniqueConstraint)
    ]
    if len(unique_constraint) == 0:
        return []
    elif len(unique_constraint) > 1:
        raise Exception(
            "'get_unique_constraint_names' can't be used for table with multiple contraints."
        )
    else:  # 1
        u = unique_constraint[0]
        return [c.name for c in u.columns]



import click
import contessa
import alembic.config
import os
import sys

from contessa.migration import MigrationsResolver
from contessa.alembic.packages_migrations import migration_map

here = os.path.dirname(os.path.abspath(__file__))
alembic_ini_path = os.path.join(here, "alembic.ini")


@click.command()
@click.option(
    "-u",
    "--url",
    help="Connection string to database in the form driver://user:pass@localhost/dbname.",
    required=True,
)
@click.option(
    "-s", "--schema", help="Schema containing tables to migrate.", required=True
)
@click.option("-v", "--version", help="Version of package to migrate.", required=True)
@click.option(
    "-f",
    "--force",
    help="""
    Force migration. Used for the package migration that differs from the current package version.""",
    is_flag=True,
)
def main(url, schema, version, force):
    if version != contessa.__version__ and not force:
        raise Exception(
            f"""
            Versions do not match. The migration is possible only to the current Contessa version.
            Use '{contessa.__version__}' as an input parameter.
            """
        )

    migration = MigrationsResolver(migration_map, version, url, schema)
    command = migration.get_migration_to_head()

    if command is None:
        print(
            f"Contessa database schema {schema} is already migrated to version {version}."
        )
        sys.exit()

    alembic_args = [
        "-x",
        f"sqlalchemy_url={url}",
        "-x",
        f"schema={schema}",
        "-c",
        alembic_ini_path,
        command[0],
        command[1],
    ]

    alembic.config.main(argv=alembic_args)


if __name__ == "__main__":
    main()
